#define input1 	2
#define input2 	3
#define picture1 	4
#define NL1 	5
#define NL2 	6
#define prestart1 	7
#define psline1 	8
#define psline2 	9
#define psline3 	10
#define elementlist1 	11
#define elementlist2 	12
#define elementlist3 	13
#define term1 	14
#define term2 	15
#define term3 	16
#define term4 	17
#define element1 	18
#define element2 	19
#define element3 	20
#define element4 	21
#define element5 	22
#define element6 	23
#define element7 	24
#define element8 	25
#define element9 	26
#define element10 	27
#define element11 	28
#define element12 	29
#define lbrace1 	30
#define namedobj1 	31
#define namedobj2 	32
#define suffix1 	33
#define suffix2 	34
#define suffix3 	35
#define position1 	36
#define position2 	37
#define position3 	38
#define position4 	39
#define assignlist1 	40
#define assignlist2 	41
#define command1 	42
#define command2 	43
#define command3 	44
#define command4 	45
#define command5 	46
#define command6 	47
#define command7 	48
#define command8 	49
#define command9 	50
#define command10 	51
#define command11 	52
#define optnl1 	53
#define optnl2 	54
#define ifpart1 	55
#define elsehead1 	56
#define for1 	57
#define for2 	58
#define stringexpr1 	59
#define stringexpr2 	60
#define string1 	61
#define string2 	62
#define assignment1 	63
#define assignment2 	64
#define assignrhs1 	65
#define assignrhs2 	66
#define expression1 	67
#define expression2 	68
#define expression3 	69
#define expression4 	70
#define expression5 	71
#define ifhead1 	72
#define setlogic1 	73
#define logexpr1 	74
#define logexpr2 	75
#define forhead1 	76
#define forincr1 	77
#define do1 	78
#define do2 	79
#define by1 	80
#define by2 	81
#define redirect1 	82
#define redirect2 	83
#define redirect3 	84
#define resetlist1 	85
#define resetlist2 	86
#define systemcmd1 	87
#define sprintf1 	88
#define sprintf2 	89
#define exprlist1 	90
#define exprlist2 	91
#define object1 	92
#define object2 	93
#define object3 	94
#define object4 	95
#define object5 	96
#define object6 	97
#define object7 	98
#define object8 	99
#define object9 	100
#define object10 	101
#define object11 	102
#define object12 	103
#define object13 	104
#define object14 	105
#define object15 	106
#define object16 	107
#define object17 	108
#define object18 	109
#define object19 	110
#define object20 	111
#define object21 	112
#define object22 	113
#define object23 	114
#define object24 	115
#define object25 	116
#define object26 	117
#define object27 	118
#define openblock1 	119
#define block1 	120
#define block2 	121
#define block3 	122
#define block4 	123
#define optexp1 	124
#define optexp2 	125
#define closeblock1 	126
#define objectwith1 	127
#define objectwith2 	128
#define objectwith3 	129
#define pair1 	130
#define pair2 	131
#define nth1 	132
#define nth2 	133
#define nth3 	134
#define primobj1 	135
#define primobj2 	136
#define primobj3 	137
#define primobj4 	138
#define shift1 	139
#define shift2 	140
#define shift3 	141
#define location1 	142
#define location2 	143
#define location3 	144
#define location4 	145
#define location5 	146
#define place1 	147
#define place2 	148
#define place3 	149
#define place4 	150
#define place5 	151
#define factor1 	152
#define factor2 	153
#define factor3 	154
#define placename1 	155
#define placename2 	156
#define placename3 	157
#define placename4 	158
#define ncount1 	159
#define ncount2 	160
#define ncount3 	161
#define logprod1 	162
#define logprod2 	163
#define logval1 	164
#define logval2 	165
#define logval3 	166
#define lcompare1 	167
#define lcompare2 	168
#define lcompare3 	169
#define lcompare4 	170
#define primary1 	171
#define primary2 	172
#define primary3 	173
#define primary4 	174
#define primary5 	175
#define primary6 	176
#define primary7 	177
#define primary8 	178
#define primary9 	179
#define primary10 	180
#define primary11 	181
#define primary12 	182
#define primary13 	183
