s%<EOF>%end of file%
s%<EMPTY>%blank%
s%<ERROR>%syntax error%
s%<float>%number%
s%<name>%identifier%
s%<Label>%label%
s%<LaTeX>%LaTeX%
s%<string>%string%
s%<arg>%$ argument%
s%<DotPS>%.PS%
s%<DotPE>%end of file or .PE%
s%";"%"end of line or ;"%
s%"th"%"ordinal: st or rd or nd or th"%
s%"\.n"%".top or .n"%
s%"\.s"%".bot or .s"%
s%"\.e"%".right or .e"%
s%"\.w"%".left or .w"%
s%<endfor>%end of for {...} contents%
s%<corner>%compass corner: .n .ne .center .end etc%
s%<compare>%logical operator%
s%<param>%attribute .ht .wid etc%
s%<func1>%function (1 arg)%
s%<func2>%function (2 args)%
s%<linetype>%line type: dotted etc%
s%<colrspec>%color spec: colored outlined shaded%
s%<textpos>%text position: ljust rjust above below center%
s%<arrowhd>%arrowhead parameter: <- -> <->%
s%<directon>%direction: up down right left%
s%<primitiv>%drawn object: box circle line etc%
s%<envvar>%environment variable%
